from .add_fact_agenda_action import AddFactAgendaAction
from .assert_fact_action import AssertFactAction
from .fact import Fact
from .periodic_action import PeriodicAction
from .psagent import PSAgent
from .retract_fact_action import RetractFactAction
from .rule import Antecedent
from .rule import Consequent
from .rule import FunctionConsequent
from .rule import Parameter
from .rule import TestAntecedent
from .rule import Rule
from .run_agenda_action import RunAgendaAction
from .sp_periodic_data import PeriodicState
from .sp_periodic_data import SPPeriodicData
from .template import Slot
from .template import Template
from .template import TypeSlot
from .update_fact_action import UpdateFactAction



from .rule import Antecedent
from ..betha_node import BethaNode
from ..alpha_node import AlphaNode
from ..rete import Rete
from abc import abstractmethod
from .run_agenda_action import RunAgendaAction
from .add_fact_agenda_action import AddFactAgendaAction
from .assert_fact_action import AssertFactAction
from .retract_fact_action import RetractFactAction
from .update_fact_action import UpdateFactAction
from ....kernel.agent.Agent import Agent

class PSAgent(Agent):
    
    def __init__(self, agentID):
        self.__rules = []
        self.engine = Rete()
        self.__templates = []
        self.__periodic_action = None
        super().__init__(agentID)

    def add_template(self, template):
        self.__templates.append(template)

    def add_rule(self, rule):
        self.__rules.append(rule)

    @abstractmethod
    def def_templates(self):
        pass

    @abstractmethod
    def def_constructs(self):
        pass

    @abstractmethod
    def set_periocic_action(self):
        pass

    def build(self):
        for template in self.__templates:
            self.engine.add_alpha_node(AlphaNode(template))
        for rule in self.__rules:
            betha_node = BethaNode()
            betha_node.name = rule.name
            for consequent in rule.get_consequents():
                betha_node.add_consequent(consequent)
            for antecedent in rule.get_antecedents():
                betha_node.add_antecedent(antecedent)
                if isinstance(antecedent, Antecedent):
                    alpha_node = self.engine.get_alpha_node(antecedent.template)
                    alpha_node.add_beta_node(betha_node)
                    betha_node.add_alpha_node(alpha_node)

    def setUp(self):
        self.state = {
            'alive': False,
            'time': None
        }
        self.addBehavior('agenda_process')
        self.bindAction('agenda_process', 'run_agenda', RunAgendaAction())
        self.bindAction('agenda_process', 'add_fact_agenda', AddFactAgendaAction())
        self.addBehavior('process')
        self.bindAction('process', 'assert_fact', AssertFactAction())
        self.bindAction('process', 'update_fact', UpdateFactAction())
        self.bindAction('process', 'retract_fact', RetractFactAction())
        self.__periodic_action = self.set_periocic_action()
        if self.__periodic_action:
            self.addBehavior('periodic_process')
            self.bindAction('periodic_process', 'periodic', self.__periodic_action)
        # Build RETE
        self.def_templates()
        self.def_constructs()
        self.build()
        # Load data RETE
        self.engine.load_agenda()
        self.engine.load_work_memory()

    def shutdown(self):
        pass



from .psagent.fact import Fact
from ...kernel.system.Adm import Adm
from .exceptions import PSException

class Rete():

    __agenda = []
    __new_agenda = []
    __alpha_dict = {}

    def add_alpha_node(self, alpha_node):
        self.__alpha_dict[alpha_node.pattern] = alpha_node;
    
    def get_alpha_node(self, pattern):
        return self.__alpha_dict[pattern]

    def assert_fact(self, fact):        
        self.__alpha_dict[fact.template].evaluate(fact, True)

    def update_fact(self, fact):        
        self.__alpha_dict[fact.template].evaluate(fact, False)

    def retract_fact(self, fact):        
        self.__alpha_dict[fact.template].retract(fact)

    def load_fact(self, fact):        
        self.__alpha_dict[fact.template].load_fact(fact)

    def add_fact_agenda(self, fact):
        alpha_node = self.__alpha_dict[fact.template]
        # Actualiza el ID del fact
        if not fact.id:
            fact.update_id(alpha_node.id)        
        # Actualiza el objeto
        updatedList = []
        for i in range(0, len(self.__agenda)):
            cur_fact = self.__agenda[i]
            if cur_fact.id == fact.id:
                for key, item in fact.obj.items():
                    cur_fact.obj[key] = item
                updatedList.append((i, cur_fact))
                break
        # Actualiza la agenda
        db = Adm().getDBConnection()
        if len(updatedList) > 0:
            for up in updatedList:
                self.__agenda.pop(up[0])
                self.__agenda.append(up[1])
                db["agenda"].delete_one({"id": up[1].id})
                db["agenda"].insert(up[1].to_dict())    
        else:
            self.__agenda.append(fact)
            db["agenda"].insert(fact.to_dict())

    def run_agenda(self):
        # Propaga los hechos
        for fact in self.__agenda:
            self.assert_fact(fact)
        # Reinicia la agenda
        db = Adm().getDBConnection()
        db["agenda"].delete_many({})
        self.__agenda = []
        
    def load_agenda(self):
        db = Adm().getDBConnection()
        if db:
            fact_list = db["agenda"].find({})
            for document in fact_list:
                fact = Fact(document['template'], document['obj'])
                fact.id = document['id']
                self.add_fact_agenda(fact)
        else:
            raise PSException('[Fatal, load_agenda]: Production system requires mondodb database.')

    def load_work_memory(self):
        db = Adm().getDBConnection()
        if db:
            fact_list = db["work_memory"].find({})
            for document in fact_list:
                fact = Fact(document['template'], document['obj'])
                fact.id = document['id']
                self.load_fact(fact)
        else:
            raise PSException('[Fatal, load_work_memory]: Production system requires mondodb database.')



from threading import Timer
from abc import abstractmethod
from ....kernel.agent.Action import Action
from .sp_periodic_data import PeriodicState, SPPeriodicData

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------
class PeriodicAction(Action):

    def __init__(self):
        self.__timer = None
        super().__init__()
    
    @abstractmethod
    def periodic_execute(self):
        pass

    def handler(self):
        """ Timeout handler """
        self.periodic_execute()
        self.adm.sendEvent(self.agent.id, 'periodic', SPPeriodicData(PeriodicState.PROCESS))

    def execute(self, data):
        """
        @param data
        """
        if isinstance(data, SPPeriodicData):
            if data.command == PeriodicState.START:
                self.agent.state['alive'] = True
                self.agent.state['time'] = data.time
                self.__timer = Timer(data.time, self.handler)
                self.__timer.start()
            elif data.command == PeriodicState.STOP:
                self.agent.state['alive'] = False
                if self.__timer:
                    self.__timer.cancel()
                    self.__timer = None
            elif data.command == PeriodicState.PROCESS:
                if self.agent.state['alive']:
                    self.__timer = Timer(self.agent.state['time'], self.handler)
                    self.__timer.start()



import uuid
import scrapy
from abc import abstractmethod
import scrapy.crawler as crawler
from .CrawlData import CrawlData
from twisted.internet import reactor
from scrapy.spiders import CrawlSpider
from .exceptions import SpiderException
from ..kernel.agent.Action import Action
from multiprocessing import Process, Queue
from scrapy.utils.log import configure_logging
from scrapy.utils.project import get_project_settings


class CrawlAction(Action, CrawlSpider):
    
    def __init__(self, cus_urls = [], cus_allowed_domains = [], agent = None):
        self.start_urls = cus_urls
        self.allowed_domains = cus_allowed_domains
        self.name = str(uuid.uuid1())
        super().__init__()
        if agent:
            self.agent = agent
            
    def execute(self, data):
        """ 
        Response.
        @param data Event data 
        """
        try:
            if isinstance(data, CrawlData):
                self.run_spider(self.agent, data)
            else:
                raise SpiderException('[Error, execute]: The object data must be instance of SpiderData')
        except Exception as e:
            raise SpiderException('[Warning, execute]: %s' % str(e))

    @abstractmethod
    def case_without_rules_trigger(self):
        pass
        
    def run_spider(self, agent, data):
        """Run the spider asocited to agent.
        @param spider:class Calss of SpiderAction.
        @param data:SpiderData dato to scrape
        """
        def f(q):
            try:
                s = get_project_settings()
                s.update({
                    "LOG_ENABLED": "False",
                    "TELNETCONSOLE_ENABLED": "False",
                    "USER_AGENT":'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36'
                })
                #configure_logging({'LOG_FORMAT': '%(levelname)s: %(message)s'})
                runner = crawler.CrawlerRunner(s)
                agent.state['fire'] = False
                agent.state['data'] = data
                deferred = runner.crawl(
                    agent.class_element, 
                    cus_urls = data.getUrls(),
                    cus_allowed_domains = data.getAllowedDomains(),
                    agent = agent
                )
                deferred.addBoth(lambda _: reactor.stop())
                reactor.run()
                if not agent.state['fire']:
                    q.put(self.case_without_rules_trigger())
                q.put(None)
            except Exception as e:
                q.put(e)
                raise SpiderException('[Warning, execute]: %s' % str(e))
        try:
            q = Queue()
            p = Process(target=f, args=(q,))
            p.start()
            print("Crawl::INIT::", p)
            result = q.get()
            print("Crawl::END::", p)
            p.join()
            if result is not None:
                data.setUrls(result)

                q = Queue()
                p = Process(target=f, args=(q,))
                p.start()
                print("(2)::Crawl::INIT::", p)
                result = q.get()
                print("(2)::Crawl::END::", p)
                p.join()
                
                #raise result
        except Exception as e:
            raise SpiderException('[Warning, execute]: %s' % str(e))



from enum import Enum 

class TypeSlot(Enum):
    KEY = 1
    PROPERTY = 2
    COUNTER = 3
    
class Slot():

    def __init__(self, name, slot_type):
        self.name = name
        self.slot_type = slot_type

class Template():

    def __init__(self, name, *slots):
        self.slots = []
        self.name = name
        for slot in slots:
            self.slots.append(slot)


from .CrawlAction import CrawlAction
from .CrawlData import CrawlData
from .exceptions import CrawlException
from .exceptions import SpiderException
from .SpiderAction import SpiderAction
from .SpiderAgent import SpiderAgent
from .SpiderData import SpiderData



from enum import Enum 

class PeriodicState(Enum):
    START = 1
    STOP = 2
    PROCESS = 3

class SPPeriodicData():
    command = None
    def __init__(self, command = PeriodicState.START, time = 60):
        self.command = command
        self.time = time



import uuid

class Fact():
    
    def __init__(self, template, obj):
        self.id = None
        self.template = template
        self.obj = obj
    
    def to_dict(self):
        dto = {}
        dto["id"] = self.id
        dto["template"] = self.template
        dto["obj"] = self.obj
        return dto

    def update_id(self, prop):
        if not self.id:
            if prop in self.obj:
                self.id = "%s_%s" % (self.template, self.obj[prop])
            else: 
                self.id = self.template

    def generate_id(self):
        self.id = str(uuid.uuid1())



class Antecedent():
    
    def __init__(self, template, template_property, value):
        self.template = template
        self.template_property = template_property
        self.value = value

class TestAntecedent():
    
    def __init__(self, operation, *parametres):
        self.parameters = []
        self.operation = operation
        for param in parametres:
            self.parameters.append(param)

class Consequent():
    
    def __init__(self, template, template_property, value):
        self.template = template
        self.template_property = template_property
        self.value = value

class Parameter():
    
    def __init__(self, template, template_property):
        self.template = template
        self.template_property = template_property

class FunctionConsequent():
    
    def __init__(self, fun, *parametres):
        self.parameters = []
        self.fun = fun
        for param in parametres:
            self.parameters.append(param)

class Rule:

    def __init__(self, name, *components):
        self.__antecedents = []
        self.__consequents = []
        self.name = name
        for component in components:
            if isinstance(component, Antecedent) or isinstance(component, TestAntecedent):
                self.__antecedents.append(component)
            else:
                self.__consequents.append(component)

    def get_antecedents(self):
        return self.__antecedents

    def get_consequents(self):
        return self.__consequents



class SpiderData():

    def __init__(self, urls = None):
        self.__urls = None
        if not urls:
            self.__urls = []
    
    def setUrls(self, urls):
        self.__urls = urls

    def addUrl(self, url):
        self.__urls.append(url)

    def getUrls(self):
        return self.__urls
